/**************************************************************************\
  class Mesh
  ==========
\**************************************************************************/

#include "Mesh.h"

#include <string>
#include <vector>
#include <list>

using namespace std;

// Constructors and destructors
//==========================================================================
Mesh::Mesh(bool inputIsExtraCompPCM):
  PointGroup(2,1), smoothTriangles(0),
  inputIsExtraCompPCM(inputIsExtraCompPCM)
{
    triangles.reserve(1000);
}
Mesh::~Mesh() {}

Mesh::Triangle::Triangle():
    p1(0),p2(0),p3(0),n1(0),n2(0),n3(0),smooth(false) {}

const string& Mesh::Name()
{
    static const string name("mesh");
    return name;
}

//==========================================================================
// Reading
//==========================================================================
// Read mesh from povray file
//==========================================================================
bool Mesh::ReadPovray(InputStream& is)
{
    degenerateTriangles=0;

    static const char* punct[5]={"{","}","<",">",","};
    static const vector<string> punctuation(punct,punct+5);
    static const char* ident[3]={"mesh","triangle","smooth_triangle"};
    static const vector<string> identifiers(ident,ident+3);

    Parser reader(is, punctuation, identifiers);
    reader.SkipStrings();

    unsigned bracketCnt=0;

    if(!Options::IgnoreMesh())
    {
      if(!reader.ReadUntil(Parser::Token(Parser::IDENTIFIER,"mesh")))
        return false;
      if(!reader.Expect(Parser::Token(Parser::PUNCTUATION,"{"),
                        "after 'mesh'"))
        return false;
      bracketCnt=1;
    }

    Parser::Token token;

    while(true)
    {
	token = reader.NextToken();

        if(token.Type() == Parser::END_OF_FILE)
        {
          if(Options::IgnoreMesh() && triangles.size()>0)
              Report();
          bool ret = Options::IgnoreMesh() && triangles.size()>0;
          if(ret && Options::Optimize())
            Optimize();
          return ret; /**** End reading (1) ****/
        }

	if(token.Type() == Parser::PUNCTUATION)
        {
            switch(token.Data()[0])
            {
              case '{':
                  bracketCnt++; break;
              case '}':
                  if(bracketCnt==0)
                  {
                      Options::PrintErrorMsg(
                        "Error: Unmatched '}' found at line %u\n",
                        is.LineNumber());
                      return false;
                  }
                  bracketCnt--;
                  if(bracketCnt==0 && !Options::IgnoreMesh())
                  {
                    Report();
                    bool ret = triangles.size()!=0;
                    if(ret && Options::Optimize())
                      Optimize();
                    return ret; /**** End reading (2) ****/
                  }
                  break;
            }
        }

        if(token.Type() == Parser::IDENTIFIER)
        {
            if(token.Data() == "triangle")
            { if(!ReadTriangle(reader)) return false; }
            else if(token.Data() == "smooth_triangle")
            { if(!ReadSmoothTriangle(reader)) return false; }
        }
    }
}

void Mesh::Report()
{
    if(triangles.size() == 0)
      Options::PrintMsg("Empty mesh block found\n");
    else
    {
      Options::PrintMsg("Found mesh with ");
      if(TriangleCount() > 0)
        Options::PrintMsg("%u triangles", TriangleCount());
      if(SmoothTriangleCount() > 0)
      { if(TriangleCount() > 0)
          Options::PrintMsg(", ");
        Options::PrintMsg("%u smooth triangles", SmoothTriangleCount());
      }
      Options::PrintMsg(" and %u points.\n", PointCount());
      if(degenerateTriangles)
        Options::PrintMsg("%u degenerate triangles ignored.\n",
                          degenerateTriangles);
    }
}

// Check if triangle is degenerate
//--------------------------------
static inline bool isDegenerate(const vector<Point>& p)
{
    return !Options::DontRemove() &&
	   (p.size()==3 ? (p[0]==p[1] || p[0]==p[2] || p[1]==p[2]) :
			  (p[0]==p[2] || p[0]==p[4] || p[2]==p[4]));
}
// Read triangle (3 vectors)
//--------------------------------------------------------------------------
bool Mesh::ReadTriangle(Parser& reader)
{
    if(!reader.Expect(Parser::Token(Parser::PUNCTUATION,"{"),
                      "after 'triangle'"))
      return false;

    static vector<Point> points(3);
    if(!ReadPoints(reader, points)) return false;

    if(!isDegenerate(points))
    {
        unsigned ind = triangles.size();
        triangles.resize(ind+1);
        triangles[ind].smooth = false;

        triangles[ind].p1 = AddPoint(points[0]);
        triangles[ind].p2 = AddPoint(points[1]);
        triangles[ind].p3 = AddPoint(points[2]);
    }
    else
      degenerateTriangles++;

    if(!reader.ReadBlock(1))
    { Options::PrintErrorMsg("Missing '}'\n"); return false; }
    return true;
}

// Read smooth triangle (6 vectors)
//--------------------------------------------------------------------------
bool Mesh::ReadSmoothTriangle(Parser& reader)
{
    if(!reader.Expect(Parser::Token(Parser::PUNCTUATION,"{"),
                      "after 'smooth_triangle'"))
      return false;

    static vector<Point> points(6);
    if(!ReadPoints(reader, points)) return false;

    if(!isDegenerate(points))
    {
        unsigned ind = triangles.size();
        triangles.resize(ind+1);
        triangles[ind].smooth = true;

        triangles[ind].p1 = AddPoint(points[0]);
        triangles[ind].n1 = AddPoint(points[1], 1);
        triangles[ind].p2 = AddPoint(points[2]);
        triangles[ind].n2 = AddPoint(points[3], 1);
        triangles[ind].p3 = AddPoint(points[4]);
        triangles[ind].n3 = AddPoint(points[5], 1);

        smoothTriangles++;
    }
    else
      degenerateTriangles++;

    if(!reader.ReadBlock(1))
    { Options::PrintErrorMsg("Missing '}'\n"); return false; }
    return true;
}

// Read mesh from PCM file
//==========================================================================
bool Mesh::ReadPCM(InputStream& is)
{
    degenerateTriangles=0;

    static const char* punct[1]={","};
    static const vector<string> punctuation(punct,punct+1);
    static const vector<string> identifiers;
    static const Parser::Token comma(Parser::PUNCTUATION,",");

    Parser reader(is, punctuation, identifiers);

    ReadPoints(reader);

    // Read number of triangles:
    unsigned tricnt = reader.NextToken().DataAsInt();
    if(!reader.Expect(comma,"")) return false;
    smoothTriangles = reader.NextToken().DataAsInt();
    if(!reader.Expect(comma,"")) return false;

    triangles.resize(tricnt+smoothTriangles);

    // Read triangles:
    bool useprev=false, invert=false;
    unsigned p1, p2=0, p3=0, n1=0, n2=0, n3=0, tmp;
    for(unsigned i=0; i<tricnt; i++)
    {
        if(inputIsExtraCompPCM)
        {
            p1 = reader.NextToken().DataAsInt();
            if(!reader.Expect(comma,"")) return false;
            if(p1==0) { useprev=invert=false; i--; continue; }
            if(useprev) { tmp=p1; p1=p2; p2=p3; p3=tmp; invert^=1; }
            else
            { p2 = reader.NextToken().DataAsInt();
              if(!reader.Expect(comma,"")) return false;
              p3 = reader.NextToken().DataAsInt();
              if(!reader.Expect(comma,"")) return false;
            }
            if(useprev && invert)
            { triangles[i].p1 = GivePointPtr(p2-1);
              triangles[i].p2 = GivePointPtr(p1-1);
              triangles[i].p3 = GivePointPtr(p3-1);
            }
            else
            { triangles[i].p1 = GivePointPtr(p1-1);
              triangles[i].p2 = GivePointPtr(p2-1);
              triangles[i].p3 = GivePointPtr(p3-1);
            }
            useprev = true;
        }
        else
        {
            triangles[i].p1 = GivePointPtr(reader.NextToken().DataAsInt());
            if(!reader.Expect(comma,"")) return false;
	    triangles[i].p2 = GivePointPtr(reader.NextToken().DataAsInt());
            if(!reader.Expect(comma,"")) return false;
	    triangles[i].p3 = GivePointPtr(reader.NextToken().DataAsInt());
            if(!reader.Expect(comma,"")) return false;
            triangles[i].smooth=false;
        }
    }

    // Read smooth triangles:
    useprev=invert=false;
    for(unsigned i=0; i<smoothTriangles; i++)
    {
        if(inputIsExtraCompPCM)
        {
            p1 = reader.NextToken().DataAsInt();
            if(!reader.Expect(comma,"")) return false;
            if(p1==0) { useprev=invert=false; i--; continue; }
            if(useprev)
            { n1 = reader.NextToken().DataAsInt();
              if(!reader.Expect(comma,"")) return false;
              tmp=p1; p1=p2; p2=p3; p3=tmp;
              tmp=n1; n1=n2; n2=n3; n3=tmp;
              invert^=1;
            }
            else
            { p2 = reader.NextToken().DataAsInt();
              if(!reader.Expect(comma,"")) return false;
              p3 = reader.NextToken().DataAsInt();
              if(!reader.Expect(comma,"")) return false;
              n1 = reader.NextToken().DataAsInt();
              if(!reader.Expect(comma,"")) return false;
              n2 = reader.NextToken().DataAsInt();
              if(!reader.Expect(comma,"")) return false;
              n3 = reader.NextToken().DataAsInt();
              if(!reader.Expect(comma,"")) return false;
            }
            if(useprev && invert)
            { triangles[i].p1 = GivePointPtr(p2-1);
              triangles[i].n1 = GivePointPtr(n2-1);
              triangles[i].p2 = GivePointPtr(p1-1);
              triangles[i].n2 = GivePointPtr(n1-1);
              triangles[i].p3 = GivePointPtr(p3-1);
              triangles[i].n3 = GivePointPtr(n3-1);
            }
            else
            { triangles[i].p1 = GivePointPtr(p1-1);
              triangles[i].n1 = GivePointPtr(n1-1);
              triangles[i].p2 = GivePointPtr(p2-1);
              triangles[i].n2 = GivePointPtr(n2-1);
              triangles[i].p3 = GivePointPtr(p3-1);
              triangles[i].n3 = GivePointPtr(n3-1);
            }
            useprev = true;
        }
        else
        {
            triangles[i].p1 = GivePointPtr(reader.NextToken().DataAsInt());
            if(!reader.Expect(comma,"")) return false;
            triangles[i].n1 = GivePointPtr(reader.NextToken().DataAsInt());
            if(!reader.Expect(comma,"")) return false;
            triangles[i].p2 = GivePointPtr(reader.NextToken().DataAsInt());
            if(!reader.Expect(comma,"")) return false;
            triangles[i].n2 = GivePointPtr(reader.NextToken().DataAsInt());
            if(!reader.Expect(comma,"")) return false;
            triangles[i].p3 = GivePointPtr(reader.NextToken().DataAsInt());
            if(!reader.Expect(comma,"")) return false;
            triangles[i].n3 = GivePointPtr(reader.NextToken().DataAsInt());
            if(!reader.Expect(comma,"")) return false;
        }
        triangles[i].smooth=true;
    }

    if(smoothTriangles>0 &&
       (Options::TransfCnt() > 0 || Options::Optimize() ||
        Options::Normalize()))
      SeparatePointsFromNormals();

    Report();
    if(Options::Optimize()) Optimize();

    return true;
}

void Mesh::SeparatePointsFromNormals()
{
    // Set index of vertices to 1
    for(unsigned i=0; i<triangles.size(); i++)
    {
	triangles[i].p1->Index(1);
	triangles[i].p1->Index(1);
	triangles[i].p1->Index(1);
    }
    // Move all normal points to group 1. If the index is 1, we have to
    // duplicate it.
    for(unsigned i=0; i<triangles.size(); i++)
    {
	if(triangles[i].smooth)
        {
	    if(triangles[i].n1->Index() == 1)
              triangles[i].n1 = AddPoint(*triangles[i].n1, 1);
            else
              MovePoint(triangles[i].n1, 0, 1);

	    if(triangles[i].n2->Index() == 1)
              triangles[i].n2 = AddPoint(*triangles[i].n2, 1);
            else
              MovePoint(triangles[i].n2, 0, 1);

	    if(triangles[i].n3->Index() == 1)
              triangles[i].n3 = AddPoint(*triangles[i].n3, 1);
            else
              MovePoint(triangles[i].n3, 0, 1);
        }
    }
    UpdateCurrentLimits();
}

//==========================================================================
// Writing
//==========================================================================
// Write in Povray format
//==========================================================================
bool Mesh::WritePovray(ostream& os)
{
    os << "mesh{" << endl;
    for(unsigned i=0; i<triangles.size(); i++)
    {
	if(!triangles[i].smooth)
        {
            os << " triangle{<"; triangles[i].p1->Print(os);
            os << "><"; triangles[i].p2->Print(os);
            os << "><"; triangles[i].p3->Print(os);
            os << ">}" << endl;
        }
        else
        {
            os << " smooth_triangle{<"; triangles[i].p1->Print(os);
            os << "><"; triangles[i].n1->Print(os);
            os << "><"; triangles[i].p2->Print(os);
            os << "><"; triangles[i].n2->Print(os);
            os << "><"; triangles[i].p3->Print(os);
            os << "><"; triangles[i].n3->Print(os);
            os << ">}" << endl;
        }
    }
    os << "}" << endl;
    return true;
}

// Write in mesh2 format
//==========================================================================
bool Mesh::WritePovray2(ostream& os)
{
    if(TriangleCount()>0)
    {
        os << "mesh2{" << endl;
        os << "vertex_vectors{";
        WritePoints(os, "<>", 0);
        os << "}" << endl << "face_indices{" << TriangleCount() << ","
           <<endl;
        unsigned cnt=0;
        for(unsigned i=0; i<triangles.size(); i++)
        {
            if(!triangles[i].smooth)
            {
                os << "<" <<  triangles[i].p1->Index()
                   << "," <<  triangles[i].p2->Index()
                   << "," <<  triangles[i].p3->Index() << ">";
                if(cnt%Options::trianglesPerLine == Options::trianglesPerLine-1)
                  os << endl;
                cnt++;
            }
        }
        os << "}}" << endl;
    }
    if(SmoothTriangleCount()>0)
    {
        SetAllPointIndex(0, 0); SetAllPointIndex(1, 0);
        SeparatePointsFromNormals();

        os << "mesh2{" << endl;
        os << "vertex_vectors{";
        WritePoints(os, "<>", 0);

	os << "}" << endl << "normal_vectors{";
        WritePoints(os, "<>", 1);

        os << "}" << endl << "face_indices{" << SmoothTriangleCount() << ","
           <<endl;
        unsigned cnt=0;
        for(unsigned i=0; i<triangles.size(); i++)
        {
            if(triangles[i].smooth)
            {
                os << "<" <<  triangles[i].p1->Index()
                   << "," <<  triangles[i].p2->Index()
                   << "," <<  triangles[i].p3->Index() << ">";
                if(cnt%Options::trianglesPerLine == Options::trianglesPerLine-1)
                  os << endl;
                cnt++;
            }
        }
        os << "}" << endl << "normal_indices{"<< SmoothTriangleCount()<<","
           << endl;
        cnt=0;
        for(unsigned i=0; i<triangles.size(); i++)
        {
            if(triangles[i].smooth)
            {
                os << "<" <<  triangles[i].n1->Index()
                   << "," <<  triangles[i].n2->Index()
                   << "," <<  triangles[i].n3->Index() << ">";
                if(cnt%Options::trianglesPerLine == Options::trianglesPerLine-1)
                  os << endl;
                cnt++;
            }
        }
        os << "}}" << endl;
    }
    return true;
}

// Write in PCM format
//==========================================================================
bool Mesh::WritePCM(ostream& os)
{
    if(Options::ExtraComp())
        return WritePCM3(os);

    os << "\"PCM1\"," << endl;
    WritePoints(os);
    os << TriangleCount() << "," << SmoothTriangleCount() << ","
       << endl;
    for(unsigned i=0; i<triangles.size(); i++)
    {
        if(!triangles[i].smooth)
        {
	    os << triangles[i].p1->Index() << ","
	       << triangles[i].p2->Index() << ","
	       << triangles[i].p3->Index() << ",";
            if(i%Options::trianglesPerLine ==
               Options::trianglesPerLine-1)
              os << endl;
        }
    }
    for(unsigned i=0; i<triangles.size(); i++)
    {
        if(triangles[i].smooth)
        {
	    os << triangles[i].p1->Index() << ","
	       << triangles[i].n1->Index() << ","
	       << triangles[i].p2->Index() << ","
	       << triangles[i].n2->Index() << ","
	       << triangles[i].p3->Index() << ","
	       << triangles[i].n3->Index() << ",";
            if(i%Options::trianglesPerLine ==
               Options::trianglesPerLine-1)
              os << endl;
        }
    }
    if(triangles.size()%Options::trianglesPerLine !=
       Options::trianglesPerLine)
      os << endl;

    return true;
}

// Write in PCM3 format
//==========================================================================
bool Mesh::WritePCM3(ostream& os)
{
    os << "\"PCM3\"," << endl;
    WritePoints(os);

    Options::PrintMsg("Calculating extra compression...\n");
    CalculateExtraCompression();

    os << TriangleCount() << "," << SmoothTriangleCount() << ","
       << endl;

    for(unsigned i=0; i<extraCompData.size(); i++)
    {
        os << extraCompData[i] << ",";
        if((i&15)==15) os << endl;
    }
    return true;
}


// Optimize triangles
//==========================================================================
void Mesh::Optimize()
{
    if(Options::TransfCnt() > 0)
    {
        Options::PrintErrorMsg(
          "Sorry, can't optimize when using transformations "
          "(see documentation)\n");
        return;
    }

    Options::PrintMsg("Optimizing...\n");

    PointPtr ptr;
    for(unsigned i=0; i<triangles.size(); i++)
    {
        if(triangles[i].smooth)
        {
            ptr = FindParallelVector(triangles[i].n1);
            if(ptr) triangles[i].n1 = ptr;
            ptr = FindParallelVector(triangles[i].n2);
            if(ptr) triangles[i].n2 = ptr;
            ptr = FindParallelVector(triangles[i].n3);
            if(ptr) triangles[i].n3 = ptr;
        }
    }
    unsigned res = RemoveUnusedPoints();
    if(res) Options::PrintMsg("%u points removed.\n", res);
    else Options::PrintMsg("No points could be removed.\n", res);
}

// Remove unused points
//==========================================================================
unsigned Mesh::RemoveUnusedPoints()
{
    SetAllPointIndex(0, 0);
    SetAllPointIndex(1, 0);
    for(unsigned i=0; i<triangles.size(); i++)
    {
        triangles[i].p1->Index(1);
        triangles[i].p2->Index(1);
        triangles[i].p3->Index(1);
        if(triangles[i].smooth)
        {
            triangles[i].n1->Index(1);
            triangles[i].n2->Index(1);
            triangles[i].n3->Index(1);
        }
    }
    return RemoveAllPointsWithIndex(0);
}


//==========================================================================
// Calculate extra compression data
//==========================================================================
inline bool Mesh::isSmoothAdjacent(Triangle* tri,
                                   unsigned sind1, unsigned sind2,
                                   unsigned adjtype)
{
    switch(adjtype)
    {
      case 1: return sind1 == tri->n1->Index() &&
                     sind2 == tri->n2->Index();
      case 2: return sind1 == tri->n1->Index() &&
                     sind2 == tri->n3->Index();
      case 3: return sind1 == tri->n2->Index() &&
                     sind2 == tri->n3->Index();
      case 4: return sind1 == tri->n2->Index() &&
                     sind2 == tri->n1->Index();
      case 5: return sind1 == tri->n3->Index() &&
                     sind2 == tri->n1->Index();
      case 6: return sind1 == tri->n3->Index() &&
                     sind2 == tri->n2->Index();
    }
    return false;
}

inline unsigned Mesh::Adjacent(Triangle* tri, unsigned ind1, unsigned ind2,
                                              unsigned sind1, unsigned sind2)
{
    unsigned ret = 0;
    if(ind1 == tri->p1->Index())
    { if(ind2 == tri->p2->Index()) ret = 1;
      if(ind2 == tri->p3->Index()) ret = 2;
    }
    if(ind1 == tri->p2->Index())
    { if(ind2 == tri->p3->Index()) ret = 3;
      if(ind2 == tri->p1->Index()) ret = 4;
    }
    if(ind1 == tri->p3->Index())
    { if(ind2 == tri->p1->Index()) ret = 5;
      if(ind2 == tri->p2->Index()) ret = 6;
    }

    if(ret>0 && tri->smooth && !isSmoothAdjacent(tri, sind1, sind2, ret))
      return 0;

    return ret;
}

inline void Mesh::NonAdjacentIndex(Triangle* tri, unsigned adjtype,
                                   unsigned& ind2, unsigned& sind2)
{
    switch(adjtype)
    {
      case 1: ind2 = tri->p3->Index(); break;
      case 2: ind2 = tri->p2->Index(); break;
      case 3: ind2 = tri->p1->Index(); break;
      case 4: ind2 = tri->p3->Index(); break;
      case 5: ind2 = tri->p2->Index(); break;
      case 6: ind2 = tri->p1->Index(); break;
    }
    if(tri->smooth)
      switch(adjtype)
      {
        case 1: sind2 = tri->n3->Index(); break;
        case 2: sind2 = tri->n2->Index(); break;
        case 3: sind2 = tri->n1->Index(); break;
        case 4: sind2 = tri->n3->Index(); break;
        case 5: sind2 = tri->n2->Index(); break;
        case 6: sind2 = tri->n1->Index(); break;
      }
}

void Mesh::CalculateExtraCompression()
{
    extraCompData.clear();

    for(unsigned type=0; type<2; type++)
    {
        unsigned tricnt = type ? SmoothTriangleCount() : TriangleCount();
        if(tricnt==0) continue;

        list<Triangle*> tribuf(tricnt);
        list<Triangle*>::iterator curr = tribuf.begin();
        Triangle* basetri;
        unsigned ind1, ind2, sind1=0, sind2=0, adjtype;

        for(unsigned i=0; i<triangles.size(); i++)
          if((type==0 && !triangles[i].smooth) ||
             (type==1 && triangles[i].smooth))
          { *curr = &triangles[i];
            curr++;
          }

        while(!tribuf.empty())
        {
            if(!extraCompData.empty())
              extraCompData.push_back(0);
            curr = tribuf.begin();
            basetri = *curr;
            extraCompData.push_back(basetri->p1->Index()+1);
            ind1 = basetri->p2->Index();
            ind2 = basetri->p3->Index();
            extraCompData.push_back(ind1+1);
            extraCompData.push_back(ind2+1);
            if(type)
            {
                extraCompData.push_back(basetri->n1->Index()+1);
                sind1 = basetri->n2->Index();
                sind2 = basetri->n3->Index();
                extraCompData.push_back(sind1+1);
                extraCompData.push_back(sind2+1);
            }
            tribuf.erase(curr);

            do
            {
                adjtype = 0;
                for(curr=tribuf.begin(); curr!=tribuf.end(); curr++)
                {
		    adjtype = Adjacent(*curr, ind1, ind2, sind1, sind2);
                    if(adjtype>0) break;
                }
                if(adjtype>0)
                {
                    basetri = *curr;
		    ind1 = ind2; sind1 = sind2;
                    NonAdjacentIndex(basetri, adjtype, ind2, sind2);
                    extraCompData.push_back(ind2+1);
                    if(type) extraCompData.push_back(sind2+1);
                    tribuf.erase(curr);
                }
            } while(adjtype>0);
        }
    }
}

