/**************************************************************************\
  class Object
  ============

  Generic object base abstract class
\**************************************************************************/

#ifndef _OBJECT_H_
#define _OBJECT_H_

#include "Point.h"
#include "InStream.h"

#include <iostream>
#include <string>

class Object
{
public:
    virtual ~Object();

    struct Limits
    { Point min, max;
      bool updated;
      Limits():updated(false) {}
      inline void Update(const Point& pnt);
    };

    enum FileType { POVRAY, POVRAY2, PCM };

    // Name of this object
    virtual const std::string& Name()=0;

    virtual bool Read(InputStream&, FileType=POVRAY);
    /* Read object from stream.
       Returns true if reading was successful, else false.
     */

    virtual bool Write(std::ostream&, FileType=PCM);
    /* Write object to stream.
       Returns true if writing was successful, else false.
     */

    virtual void RoundValues()=0;

    bool ApplyTransformations(const std::vector<std::string>& transformations);

    inline const Limits& GetLimits() { return currentLimits; }

    // Transformations. Return true if successful, else false.
    virtual bool Scale(double, bool verbose=true)=0;
    virtual bool Translate(const Point&, bool verbose=true)=0;
    virtual bool SwapCoords(char axis1, char axis2, bool verbose=true)=0;
    virtual bool Rotate(double angle, const Point&,
                        const Point&, bool verbose=true)=0;

    // Default transformations:
    virtual bool Center(bool bottom,
                        bool verbose = true,
                        const Limits* = 0);
    virtual bool Scale2Bound(double size,
                             bool verbose = true,
                             const Limits* = 0);
    virtual bool PrintBoundaries(const char* caption="Limits",
                                 const Limits* = 0);

protected:
    virtual bool ReadPovray(InputStream&);
    virtual bool ReadPovray2(InputStream&);
    virtual bool ReadPCM(InputStream&);
    virtual bool WritePovray(std::ostream&);
    virtual bool WritePovray2(std::ostream&);
    virtual bool WritePCM(std::ostream&);

    // Limits of the object
    virtual void UpdateCurrentLimits()=0;
    Limits currentLimits;
};

inline void Object::Limits::Update(const Point& pnt)
{
    if(!updated) { min = max = pnt; updated = true; }
    else
    {
        if(pnt.x() < min.x())      min.x(pnt.x());
        else if(pnt.x() > max.x()) max.x(pnt.x());
        if(pnt.y() < min.y())      min.y(pnt.y());
        else if(pnt.y() > max.y()) max.y(pnt.y());
        if(pnt.z() < min.z())      min.z(pnt.z());
        else if(pnt.z() > max.z()) max.z(pnt.z());
    }
}

#endif
