/**************************************************************************\
  class Parser
  ============

  General token parser
\**************************************************************************/

#ifndef _PARSER_H_
#define _PARSER_H_

#include "Point.h"
#include "InStream.h"
#include "Options.h"

#include <iostream>
#include <string>
#include <vector>
#include <set>
#include <cstdlib>
using namespace std;

class Parser
{
public:
// Constructor:
//--------------
    Parser(InputStream&,
           const vector<string>& punctuation,
           const vector<string>& identifiers);
    /* istream: input stream
       punctuation: Punctuation characters, eg "{", "}", "<", ">"
       identifiers: Keywords. Must contain only characters 0-9, A-Z, a-Z and _
                    Can't begin with number.
     */

// Token definitions:
//--------------------
    enum token_type { NONE, IDENTIFIER, PUNCTUATION, FLOAT_NUMBER, STRING,
                      END_OF_FILE };

    class Token
    {
    public:
        inline Token(token_type t = NONE, const string& d = ""):
            type(t), data(d) {}

        inline Token& Set(token_type t, const string& d = "")
        { type=t; data=d; return *this; }

        inline token_type Type() const { return type; }
        inline const string& Data() const { return data; }
        inline double DataAsDouble() const { return atof(data.c_str()); }
        inline int DataAsInt() const { return atoi(data.c_str()); }

        inline bool operator==(const Token& t) const
        { return type==t.type && data==t.data; }
        inline bool operator!=(const Token& t) const
        { return type!=t.type || data!=t.data; }
        inline bool operator<(const Token& t) const
        { return type!=t.type ? type<t.type : data<t.data; }

    private:
        token_type type;
        string data;
    };

// Methods:
//----------
    Token NextToken();
    // Returns the next token from the stream,

    inline const Token& PeekToken();
    // Returns reference to the next token in the stream without removing it.

    bool ReadUntil(const Token& token);
    // Read tokens from stream until the given token is found.
    // Ret. value: true=token found; false=token not found (EOF reached)

    inline bool Expect(const Token& token, const char* msg);
    // Read next token and if it's not equal to the given one, print:
    // Error at line <line>: <token.Data()> expected <msg>

    inline void SkipNextTokenIf(const Token& token);
    // If the next token in the stream is equal to the given token, skip it

    bool ReadPoint(Point&, bool expand=false);
    // Read a 3D-point from the stream into the given Point.
    // Returns true if the next tokens formed a 3D-point, else false.
    // A 3D-point is a group of 3 floats separated with whitespaces
    //  and/or ",".
    // Requires that "," is one of the punctuation strings.
    // When false, numbers read so far from the stream are not pushed
    //  back. The numbers read so far are stored into the given Point. The
    //  Index() of the Point is set to the amount of numbers read (ie. 0, 1
    //  or 2); if the reading was successful, the Index() is not changed.
    //  The token which ends the reading is not removed from the stream.
    // However, if 'expand' is true and the ending token is ">", the missing
    //  values are expanded to 0. Requires that ">" is one of the punctuation
    //  strings. The ">" is not removed from the stream.

    inline void SkipStrings(bool b=true) { skipStrings=b; }
    // If set to true, NextToken() just skips strings instead of reading
    //  them into a Token and returning it.
    // This may speed up a little the reading when not interested in strings.

    bool ReadBlock(unsigned bracketCnt,
                   char openBracket='{', char closeBracket='}');
    // Read a block out of the stream. The 1st parameter is the current level
    // of block nesting. The next parameters are the opening and the closing
    // Block delimiters.

    inline const InputStream& Stream() { return is; }


//======================================================================
private:
    InputStream& is;
    vector<string> punctuation;
    const set<string> identifiers;
    bool skipStrings;
    Token nextToken;
};

inline const Parser::Token& Parser::PeekToken()
{
    if(nextToken.Type() == NONE)
      nextToken = NextToken();
    return nextToken;
}

inline bool Parser::Expect(const Token& token, const char* msg)
{
    if(NextToken() != token)
    {
        Options::PrintErrorMsg(
            "Error at line %u: '%s' expected %s\n",
            is.LineNumber(), token.Data().c_str(), msg);
        return false;
    }
    return true;
}

inline void Parser::SkipNextTokenIf(const Token& token)
{
    if(token == PeekToken()) NextToken();
}

#endif

