/**************************************************************************\
  class Point
  ===========

  Generic 3D-point class
\**************************************************************************/

#ifndef _POINT_H_
#define _POINT_H_

#include "Options.h"

#include <iostream>
#include <cmath>

class Point
{
public:
    // Constructors
    inline Point(): xCoord(0), yCoord(0), zCoord(0), index(0) {}
    inline Point(double X, double Y, double Z, unsigned index=0):
        xCoord(X), yCoord(Y), zCoord(Z), index(index) {}
    inline Point(const Point& p) { *this = p; }

    // Setting and reading
    inline void Set(double X, double Y, double Z);

    inline void x(double d) { xCoord = d; }
    inline double x() const { return xCoord; }
    inline void y(double d) { yCoord = d; }
    inline double y() const { return yCoord; }
    inline void z(double d) { zCoord = d; }
    inline double z() const { return zCoord; }

    inline void Index(unsigned i) { index=i; }
    inline unsigned Index() const { return index; }

    // Transformations
    inline void Scale(double);
    inline void Translate(const Point&);
    inline void SwapCoords(char axis1, char axis2);
    inline void NegCoord(char axis);
    void Rotate(double angle, const Point& p1, const Point& p2,
                bool translate);

    inline void SetLength(double length);
    inline void Transform(double(*TransformFunc)(double));

    // Operators
    inline bool operator==(const Point& pnt) const;
    inline bool operator!=(const Point& pnt) const;
    inline bool operator<(const Point& pnt) const;
    inline bool Compare(const Point& pnt, double tolerance);

    // Print
    inline void Print(std::ostream&);



//==========================================================================
private:
    double xCoord, yCoord, zCoord;
    unsigned index;
};

inline void Point::Set(double X, double Y, double Z)
{
    xCoord = X; yCoord = Y; zCoord = Z;
}

// Operators:
inline bool Point::operator==(const Point& pnt) const
{
    return Options::Compare(xCoord,pnt.x()) &&
           Options::Compare(yCoord,pnt.y()) &&
           Options::Compare(zCoord,pnt.z());
}
inline bool Point::operator!=(const Point& pnt) const
{
    return !(*this==pnt);
}
inline bool Point::operator<(const Point& pnt) const
{
    return Options::Compare(xCoord,pnt.x()) ?
                    ( Options::Compare(yCoord,pnt.y()) ?
                      ( zCoord<pnt.z() ):
                      ( yCoord<pnt.y() )
                    ):
                    ( xCoord<pnt.x() );
}

struct PointLess
{ inline bool operator()(const Point* x, const Point* y) const
  { return *x < *y; }
};

inline bool Point::Compare(const Point& pnt, double tolerance)
{
    return fabs(xCoord-pnt.x()) <= tolerance &&
	   fabs(yCoord-pnt.y()) <= tolerance &&
	   fabs(zCoord-pnt.z()) <= tolerance;
}

// Transformations:
inline void Point::Scale(double f)
{
    xCoord*=f; yCoord*=f; zCoord*=f;
}
inline void Point::Translate(const Point& p)
{
    xCoord+=p.x(); yCoord+=p.y(); zCoord+=p.z();
}
inline void Point::SwapCoords(char axis1, char axis2)
{
    double& c1 = (axis1=='x') ? xCoord : ((axis1=='y') ? yCoord : zCoord);
    double& c2 = (axis2=='x') ? xCoord : ((axis2=='y') ? yCoord : zCoord);
    double tmp=c1; c1=c2; c2=tmp;
}
inline void Point::NegCoord(char axis)
{
    double& c = (axis=='x') ? xCoord : ((axis=='y') ? yCoord : zCoord);
    c = -c;
}

inline void Point::SetLength(double length)
{
    double l=sqrt(xCoord*xCoord+yCoord*yCoord+zCoord*zCoord);
    if(l>0) l=length/l;
    xCoord*=l; yCoord*=l; zCoord*=l;
}

inline void Point::Transform(double(*TransformFunc)(double))
{
    xCoord = TransformFunc(xCoord);
    yCoord = TransformFunc(yCoord);
    zCoord = TransformFunc(zCoord);
}

// Print:
inline void Point::Print(std::ostream& os)
{
    os << xCoord << "," << yCoord << "," << zCoord;
}

#endif

