/**************************************************************************\
  class PointGroup
  ================
\**************************************************************************/

#include "PointGrp.h"
#include "Options.h"

#include <cmath>

// Constructor and destructor:
//========================================================================
PointGroup::PointGroup(unsigned groups, unsigned normalgroup):
    normalgroup(normalgroup), pointCnt(0)
{
    points.resize(groups);
}

PointGroup::~PointGroup()
{
    for(unsigned group=0; group<points.size(); group++)
        for(iterator iter=begin(group); iter!=end(group); iter++)
            delete *iter;
}

// Add point to group
//========================================================================
PointGroup::PointPtr PointGroup::AddPoint(const Point& pnt, unsigned group)
{
    Point* p = new Point(pnt);
    if(Options::Round()) p->Transform(&Options::Round);

    iterator i = points[group].find(p);
    if(i != points[group].end())
    {
        delete p;
        return *i;
    }

    pointCnt++;
    if(group != normalgroup) currentLimits.Update(pnt);
    points[group].insert(p);
    return p;
}

// Read points in PCM format from given	Parser
//========================================================================
bool PointGroup::ReadPoints(Parser& reader, unsigned group)
{
    Point pnt;

    unsigned count = reader.NextToken().DataAsInt();
    pointPtrs.resize(count);

    if(!reader.Expect(Parser::Token(Parser::PUNCTUATION,","),""))
      return false;
    for(unsigned i=0; i<count; i++)
    {
        reader.ReadPoint(pnt);
        pointPtrs[i] = AddPoint(pnt, group);
        if(!reader.Expect(Parser::Token(Parser::PUNCTUATION,","),""))
          return false;
    }
    return true;
}

// Write points in PCM format to given ostream
//========================================================================
void PointGroup::WritePoints(ostream& os, const char* delimiters,
                             unsigned grp)
{
    unsigned index=0;

    if(grp==(unsigned)~0)
      os << PointCount() << "," << endl;
    else
      os << points[grp].size() << "," << endl;

    for(unsigned group=(grp==(unsigned)~0?0:grp); group<points.size(); group++)
    {
	for(iterator i=begin(group); i!=end(group); i++)
        {
            if(delimiters) os << delimiters[0];
            (*i)->Print(os);
            os << (delimiters ? &delimiters[1] : ",");
            (*i)->Index(index);
            index++;
            if(index%Options::pointsPerLine == 0)  os << endl;
        }
        if(grp!=(unsigned)~0) break;
    }
    if(index%Options::pointsPerLine != 0)  os << endl;
}

// Round all values
//========================================================================
void PointGroup::RoundValues()
{
    for(unsigned group = 0; group < points.size(); group++)
      for(iterator i = begin(group); i != end(group); i++)
        (*i)->Transform(&Options::Round);
}

// Normalize all vectors in given group
//========================================================================
void PointGroup::Normalize(unsigned group)
{
    for(iterator i = begin(group); i != end(group); i++)
      (*i)->SetLength(Options::NormalLength());
}

// Update limits
//========================================================================
void PointGroup::UpdateCurrentLimits()
{
    currentLimits.updated = false;

    for(unsigned group = 0; group < points.size(); group++)
    {
        if(group == normalgroup) continue;
        for(iterator i = begin(group); i != end(group); i++)
        {
            currentLimits.Update(*(*i));
        }
    }
}

// Transformations
//========================================================================
bool PointGroup::Scale(double f, bool verbose)
{
    if(verbose)
      Options::PrintMsg("Scaling %s by %g\n", Name().c_str(), f);

    for(unsigned group = 0; group < points.size(); group++)
    {
        if(group == normalgroup) continue;
        for(iterator i = begin(group); i != end(group); i++)
          (*i)->Scale(f);
    }
    currentLimits.min.Scale(f);
    currentLimits.max.Scale(f);
    return true;
}

bool PointGroup::Translate(const Point& coords, bool verbose)
{
    if(verbose)
      Options::PrintMsg("Translating %s by <%g,%g,%g>\n",
                        Name().c_str(), coords.x(), coords.y(), coords.z());

    for(unsigned group = 0; group < points.size(); group++)
    {
        if(group == normalgroup) continue;
        for(iterator i = begin(group); i != end(group); i++)
          (*i)->Translate(coords);
    }
    currentLimits.min.Translate(coords);
    currentLimits.max.Translate(coords);
    return true;
}

bool PointGroup::SwapCoords(char axis1, char axis2, bool verbose)
{
    if(axis1 == axis2)
    {
        if(verbose)
          Options::PrintMsg("Mirroring %c coordinate\n", axis1);
        for(unsigned group = 0; group < points.size(); group++)
          for(iterator i = begin(group); i != end(group); i++)
            (*i)->NegCoord(axis1);
        currentLimits.min.NegCoord(axis1);
        currentLimits.max.NegCoord(axis1);
    }
    else
    {
        if(verbose)
          Options::PrintMsg("Swapping %c and %c coordinates\n", axis1, axis2);
        for(unsigned group = 0; group < points.size(); group++)
          for(iterator i = begin(group); i != end(group); i++)
            (*i)->SwapCoords(axis1, axis2);
        currentLimits.min.SwapCoords(axis1, axis2);
        currentLimits.max.SwapCoords(axis1, axis2);
    }
    return true;
}

bool PointGroup::Rotate(double angle, const Point& p1, const Point& p2,
                      bool verbose)
{
    if(p1 == p2)
    {
        Options::PrintMsg("Points are equal. No rotation made.\n");
        return false;
    }

    if(verbose)
      Options::PrintMsg("Rotating %s %g radians around <%g,%g,%g>-<%g,%g,%g>\n",
                        Name().c_str(),angle,
                        p1.x(),p1.y(),p1.z(),p2.x(),p2.y(),p2.z());

    for(unsigned group = 0; group < points.size(); group++)
      for(iterator i = begin(group); i != end(group); i++)
        (*i)->Rotate(angle, p1, p2, group!=normalgroup);

    UpdateCurrentLimits();
    return true;
}

// Read given number of points
//----------------------------
bool PointGroup::ReadPoints(Parser& reader, vector<Point>& points)
{
    Parser::Token token;
    for(unsigned i=0; i<points.size(); i++)
    {
	if(!reader.Expect(Parser::Token(Parser::PUNCTUATION, "<"), ""))
          return false;
        if(!reader.ReadPoint(points[i], true))
        { Options::PrintMsg("Invalid vector at line %u\n",
                            reader.Stream().LineNumber());
          return false;
        }
        if(Options::Round()) points[i].Transform(&Options::Round);
        if(!reader.Expect(Parser::Token(Parser::PUNCTUATION, ">"), ""))
          return false;
        reader.SkipNextTokenIf(Parser::Token(Parser::PUNCTUATION,","));
    }
    return true;
}

// Set the index of all points in given group to given value
//========================================================================
void PointGroup::SetAllPointIndex(unsigned group, unsigned value)
{
    for(iterator iter=begin(group); iter!=end(group); iter++)
      (*iter)->Index(value);
}

// Remove all points which have the given index
//========================================================================
unsigned PointGroup::RemoveAllPointsWithIndex(unsigned ind)
{
    unsigned cnt=0;

    for(unsigned group=0; group<points.size(); group++)
    {
	for(iterator iter=begin(group); iter!=end(group);)
        {
            if((*iter)->Index() == ind)
            {
                iterator curr=iter; iter++; // hack for djgpp
                delete (*curr);
		points[group].erase(curr);
                pointCnt--;
                cnt++;
            }
            else iter++;
        }
    }
    return cnt;
}

// Find a vector parallel to given one
//========================================================================
PointGroup::PointPtr PointGroup::FindParallelVector(PointPtr pntptr)
{
    Point pnt(*pntptr);
    pnt.SetLength(10);
    Point current;

    for(unsigned group=0; group<points.size(); group++)
    {
	for(iterator iter=begin(group); iter!=end(group); iter++)
        {
            if((*iter) == pntptr) return 0;
            current = *(*iter);
            current.SetLength(10);
            if(pnt.Compare(current, Options::ErrorTolerance()))
              return *iter;
        }
    }
    return 0;
}

