/**************************************************************************\
  class PointGroup
  ================

  Object-type class which contains (unique) points
\**************************************************************************/

#ifndef _POINTGROUP_HH_
#define _POINTGROUP_HH_

#include "Point.h"
#include "Object.h"
#include "Parser.h"

#include <set>
#include <vector>

class PointGroup : public Object
{
public:
    // Constructor and destructor:
    PointGroup(unsigned groups=1, unsigned normalgroup=(unsigned)(~0));
    /* groups: How many point groups are to be used.
       normalgroup: This group is a group of normal vectors. Certain
                    transformations are not applied to them and they don't
                    affect the limits.
     */
    virtual ~PointGroup();

    // Inherited methods:
    void RoundValues();

protected:
    void Normalize(unsigned group);

    // Types and iterators:
    typedef Point* PointPtr;
    typedef std::set<PointPtr, PointLess> plist_t;
    typedef plist_t::iterator iterator;

    inline iterator begin(unsigned group) { return points[group].begin(); }
    inline iterator end(unsigned group) { return points[group].end(); }

    // Add point to group. Returns pointer to the new point.
    PointPtr AddPoint(const Point&, unsigned group=0);
    // Move point from one group to another:
    inline void MovePoint(PointPtr, unsigned srcgroup, unsigned destgroup);

    inline unsigned PointCount() { return pointCnt; }

    void UpdateCurrentLimits();

    // Read points in PCM format from given Parser:
    bool ReadPoints(Parser&, unsigned group=0);
    // Give PointPtr to point number 'n' (when reading PCM)
    inline PointPtr GivePointPtr(unsigned n) { return pointPtrs[n]; }

    // Write points in PCM format to given ostream:
    void WritePoints(ostream&, const char* delimiters=0,
                     unsigned group=~0);

    // Read points to the given vector:
    bool ReadPoints(Parser&, vector<Point>&);

    // Find a vector parallel to given one:
    PointPtr FindParallelVector(PointPtr);

    void SetAllPointIndex(unsigned group, unsigned value);
    unsigned RemoveAllPointsWithIndex(unsigned ind);

private:
    // Own methods and variables:
    std::vector<plist_t> points;
    unsigned normalgroup, pointCnt;
    std::vector<PointPtr> pointPtrs;

    // Inherited methods:
    bool Scale(double, bool verbose=true);
    bool Translate(const Point&, bool verbose=true);
    bool SwapCoords(char axis1, char axis2, bool verbose=true);
    bool Rotate(double angle, const Point&,
                const Point&, bool verbose=true);
};

inline void PointGroup::MovePoint(PointPtr pnt,
                                  unsigned srcgroup, unsigned destgroup)
{
    points[srcgroup].erase(pnt);
    points[destgroup].insert(pnt);
}

#endif
