/**************************************************************************\
  class Union
  ===========

\**************************************************************************/

#include "Union.h"

#include "Mesh.h"
#include "BPatches.h"
#include "Parser.h"

#include <cstdio>

using namespace std;

// Constructor and destructor
//=========================================================================
Union::Union() {}

Union::~Union()
{
    for(olist_t::iterator iter=objectList.begin();
        iter!=objectList.end(); iter++)
      delete *iter;
}

const string& Union::Name()
{
    static const string name("union");
    return name;
}

// Reading
//=========================================================================
bool Union::ReadPovray(InputStream& is)
{
    Object* obj;

    if(Options::InputIsBP())
    {
	obj = new BPatches;
        if(!obj->Read(is, POVRAY))
        {
            delete obj;
            return false;
        }
        objectList.insert(objectList.end(), obj);
    }
    else
    {
        bool ret;
	do
        {
            obj = new Mesh;
	    ret = obj->Read(is, POVRAY);
            if(ret)
              objectList.insert(objectList.end(), obj);
            else
              delete obj;
        } while(ret);
    }

    if(objectList.empty())
      Options::PrintErrorMsg("No objects found\n");

    UpdateCurrentLimits();

    return !objectList.empty();
}

bool Union::ReadPCM(InputStream& is)
{
    static const char* punct[1]={","};
    static const vector<string> punctuation(punct,punct+1);
    static const vector<string> identifiers;
    Parser reader(is, punctuation, identifiers);
    Parser::Token token;
    Object* obj;

    while(true)
    {
        do
        {
            token = reader.NextToken();
            if(token.Type() == Parser::END_OF_FILE) break;
        } while(token.Type() != Parser::STRING);

        if(token.Type() == Parser::END_OF_FILE) break;
        reader.SkipNextTokenIf(Parser::Token(Parser::PUNCTUATION,","));

        obj = 0;
        if(token.Data() == "PCM1")
          obj = new Mesh();
        else if(token.Data() == "PCM3")
          obj = new Mesh(true);
        else if(token.Data() == "PCM2")
          obj = new BPatches();

        if(obj)
        {
          if(obj->Read(is, PCM))
          {
              objectList.insert(objectList.end(), obj);
          }
          else
          {
              delete obj;
              break;
          }
        }
    }

    if(objectList.empty())
      Options::PrintErrorMsg("No objects found\n");

    UpdateCurrentLimits();

    return !objectList.empty();
}

// Writing
//=========================================================================
bool Union::Write(ostream& os, FileType type)
{
    bool ret = true;

    for(olist_t::iterator iter=objectList.begin();
        iter!=objectList.end(); iter++)
    {
      if(!(*iter)->Write(os, type)) ret=false;
    }

    return ret;
}

// Transformations
//=========================================================================
void Union::RoundValues()
{
    for(olist_t::iterator iter=objectList.begin();
        iter!=objectList.end(); iter++)
      (*iter)->RoundValues();
}

bool Union::Scale(double f, bool verbose)
{
    if(verbose)
      Options::PrintMsg("Scaling objects by "DOUBLE"\n", f);

    for(olist_t::iterator iter=objectList.begin();
        iter!=objectList.end(); iter++)
      if(!(*iter)->Scale(f, false))
        return false;

    UpdateCurrentLimits();
    return true;
}

bool Union::Translate(const Point& coords, bool verbose)
{
    if(verbose)
      Options::PrintMsg("Translating objects by <"DOUBLE","DOUBLE","DOUBLE">\n",
                        coords.x(), coords.y(), coords.z());

    for(olist_t::iterator iter=objectList.begin();
        iter!=objectList.end(); iter++)
      if(!(*iter)->Translate(coords, false))
        return false;

    UpdateCurrentLimits();
    return true;
}

bool Union::SwapCoords(char axis1, char axis2, bool verbose)
{
    if(verbose)
        if(axis1 == axis2)
          Options::PrintMsg("Mirroring %c coordinate of objects\n", axis1);
        else
          Options::PrintMsg("Swapping %c and %c coordinates of objects\n",
                            axis1, axis2);

    for(olist_t::iterator iter=objectList.begin();
        iter!=objectList.end(); iter++)
      if(!(*iter)->SwapCoords(axis1, axis2, false))
        return false;

    UpdateCurrentLimits();
    return true;
}

bool Union::Rotate(double angle, const Point& p1,
                   const Point& p2, bool verbose)
{
    if(verbose)
      Options::PrintMsg("Rotating objects "DOUBLE" radians around "
                        "<"DOUBLE","DOUBLE","DOUBLE">-"
                        "<"DOUBLE","DOUBLE","DOUBLE">\n",
                        angle,p1.x(),p1.y(),p1.z(),p2.x(),p2.y(),p2.z());

    for(olist_t::iterator iter=objectList.begin();
        iter!=objectList.end(); iter++)
      if(!(*iter)->Rotate(angle, p1, p2, false))
        return false;

    UpdateCurrentLimits();
    return true;
}

bool Union::Center(bool bottom, bool verbose, const Limits* l)
{
    if(!Options::NoMeshUnion())
    {
        return Object::Center(bottom, verbose, l);
    }

    for(olist_t::iterator iter=objectList.begin();
        iter!=objectList.end(); iter++)
    {
        if(!(*iter)->Center(bottom, verbose))
          return false;
    }

    UpdateCurrentLimits();
    return true;
}

bool Union::Scale2Bound(double size, bool verbose, const Limits* l)
{
    if(!Options::NoMeshUnion())
    {
        return Object::Scale2Bound(size, verbose, l);
    }

    for(olist_t::iterator iter=objectList.begin();
        iter!=objectList.end(); iter++)
    {
        if(!(*iter)->Scale2Bound(size, verbose))
          return false;
    }

    UpdateCurrentLimits();
    return true;
}

bool Union::PrintBoundaries(const char* caption, const Limits* l)
{
    static char subCaption[] = "Limits of sub-object 00000000000000000000";

    if(!Options::NoMeshUnion())
      Object::PrintBoundaries(caption, l);

    if(Options::NoMeshUnion() || Options::SubMeshBoundaries())
    {
        unsigned counter=1;
        for(olist_t::iterator iter=objectList.begin();
            iter!=objectList.end(); iter++, counter++)
        {
            sprintf(&subCaption[21], "%u", counter);
            (*iter)->PrintBoundaries(subCaption);
        }
    }

    return true;
}


// Update limits:
//---------------
void Union::UpdateCurrentLimits()
{
    currentLimits.updated=false;
    for(olist_t::iterator iter=objectList.begin();
        iter!=objectList.end(); iter++)
    {
      currentLimits.Update((*iter)->GetLimits().min);
      currentLimits.Update((*iter)->GetLimits().max);
    }
}

